;------------------------------------------------------------------------
; HUB_EP0.ASM
; Sub-Module to HUB.ASM
;------------------------------------------------------------------------

; include related sub-modules
    INCLUDE "hub_ep0s.asm"              ; Standard device request routines
    INCLUDE "hub_ep0c.asm"              ; Hub-Class device request routines


;------------------------------------------------------------------------
TASK_HUB_EP0:
; Hub default pipe request processor (main loop task)
; Frequent requests should be prioritized to the top of the tables
; MUST re-enable EPB0 interrupt upon completion
;
; Note: the EPB0 interrupt routine filters all OUT packets so that we
; should only have to run this task when SETUP or IN packets are received.
;------------------------------------------------------------------------
    read    REG_GLOBAL_INT_ENABLE       ; test if 1ms Int is enabled
    AND     A, MASK_1024US
    JNZ     TH_EP0_TIMER_SET            ; skip if set

    enable_1ms_timer                    ; else enable 1ms int on first EP0 int
                                        ; Must have bus activity if we got this far!

TH_EP0_TIMER_SET:
    MOV     A, [HUB_EP0_TOKEN]          ; get the received token
	AND	A, SETUP_TOKEN 		            ; isolate setup
    CMP     A, SETUP_TOKEN              ; was it a setup token?
    JNZ     HUB_EP0_IN                  ; if not, must be an IN token

HUB_EP0_SETUP:
    ;following line strictly not necessary but helps CY3652 development system
    ;run more reliably with OHCI hubs
    unlock  REG_EP_B0_MODE, NAK_IN_OUT
    MOV     A, 0                        ; clear index
    MOV     X, A                        ; use X as the index
LOOK_TYPE:
    INDEX   TABLE_REQUEST_TYPE          ; fetch request type
    CMP     A, [HUB_bmRequestType]      ; = bmRequestType? (match)
    JZ      TYPE_FOUND                  ; yes
    CMP     A, INVALID_TYPE             ; end of table?
    JZ      TYPE_NOT_FOUND              ; exit
    INC     X                           ; else bump index
    MOV     A, X                        ; A = index
    JMP     LOOK_TYPE                   ; look at next type for match
TYPE_NOT_FOUND:
    JMP     HUB_EP0_STALL               ; invalid bmRequestType
TYPE_FOUND:
    MOV     A, X                        ; addr*2
    ASL                                 ; *2
    JACC    JMP_TABLE_REQUEST_TYPE      ; jump to request type processor

; return point(s) from ALL above PROC_'s (either stall or task end)
HUB_EP0_STALL:                          ; stall unsupported functions
    unlock  REG_EP_B0_MODE, STALL
    JMP     TASK_HUB_EP0_END

HUB_EP0_IN:
; Data pending for send to host
; start the next transfer
    CALL    SEND_BUFFER                 ; load the FIFO
    read    REG_EP_B0_COUNT             ; get current data 0/1 bit setting
    AND     A, MASK_DATATOGGLE          ; keep data 0/1 bit
    XOR     A, MASK_DATATOGGLE          ; toggle the data 0/1 bit
    JMP     HUB_ACK_COUNT


; No data to send back to host
NO_DATA_CONTROL:
    unlock  REG_EP_B0_COUNT, MASK_DATATOGGLE
    unlock  REG_EP_B0_MODE, STATUS_IN_ONLY
    MOV     A,0                         ; clear hub data count
    MOV     [HUB_DATA_COUNT],A
    JMP     TASK_HUB_EP0_END

; ACK next IN with count register set to HUB_BYTE_COUNT (ONLY fifo is loaded)
HUB_ACK_IN_WITH_COUNT:
    MOV     A,0                         ; clear hub data count
    MOV     [HUB_DATA_COUNT],A
HUB_ACK_IN_WITH_DATA:
    MOV     A, MASK_DATATOGGLE          ; set data 0/1 bit
HUB_ACK_COUNT:
    OR      A, [HUB_BYTE_COUNT]         ; write correct byte count
    MOV     [TEMP], A
    unlock  REG_EP_B0_COUNT, [TEMP]
    unlock  REG_EP_B0_MODE, ACK_IN_STATUS_OUT

TASK_HUB_EP0_END:
    clr_flag FLAG_HUB_EP0               ; clear the flag
    enable_hub_ep0_int                  ; re-enable the HUB EP0 interrupt
    RET

; NOTE: See JMP_TBL.ASM for JMP tables and Lookup tables for this module
; END HUB_EP0.ASM -------------------------------------------------------