;------------------------------------------------------------------------
; SUSPEND.ASM
;------------------------------------------------------------------------
; Hub suspend control routines and TASK's
;------------------------------------------------------------------------


;------------------------------------------------------------------------
TASK_HUB_SUSPEND:
; Global Suspend
; (1) r.e. [2] Page 235 11.5 Suspend and Resume
; (2) r.e. [1] Chapter 9, USB Device Framework and Figure 9-1 Device State Diagram
; FLAG_HUB_SUSPEND is set by the 1.024 ms timer interrupt when the root port
; has been idle for >3ms.
; A Suspended Hub MUST:
;   1) float all output drivers
;   2) maintain static values of all control/status bits
;   3) preserve downstream port states
; A device can be suspended from any state once powered (2).
;
; Conditions on SUSPEND --------------------------------------------------
; [2]11.5.3 Hub behavior during suspend and resume (pg 243-246)
; Root Port Idle
;       0-3ms   Suspend is pending, do nothing (have not entered this routine).
;               Hub operation normal including continued sending of resume (K)
;               on a suspended DS port (STATE_RESUME_PENDING = START or OVERRIDE ).
;       3-5ms   Suspend detected (enter this routine) FLAG_HUB_SUSPEND is set.
;               Do nothing until after 5ms.
;        >5ms   Before transitioning to the Low-Power (suspend) mode test
;               the following: (If any are true DO NOT suspend but send
;               resume(K) signaling US and all enabled DS ports and if
;               initiated by a suspended DS port, to that port also)
;               CASE 1) Resume pending (Suspended DS port previously initiated resume)
;                   STATE_RESUME_PENDING = START or OVERRIDE
;               CASE 2) C_PORT_SUSPEND bit set on any enabled port and has not been
;                   cleared by the host
;               CASE 3) C_PORT_CONNECTION (connect/disconnect) bit is still set
;                   and not cleared by the host (only if hub is Remote Wakeup)
;
; Conditions on RESUME ---------------------------------------------------
; (From Low-Power suspend mode)
; Resume can occur given one of the following events (resume at PC+1):
; CASE 1) Resume (K) on the US (root) port.
;           Hub hardware drives resume (K) downstream to all enabled (unsuspended)
;           ports. Firmware must terminate the HW forcing by 2 I/O writes to the
;           port control registers.
;           NOTE: SE0 on the root port should be detected as a RESET event.
; CASE 2) Resume (K) on an enabled or suspended DS port (dp_wakeup).
;           Hub hardware reflects resume (K) upstream and downstream to all enabled
;           (unsuspended) ports until 2 I/O writes to the port control registers.
;           A call to SEND_GLOBAL_RESUME is made in this case.
; CASE 3) If hub is DEVICE_REMOTE_WAKEUP, a DS connect or disconnect event.
;          (potential resume)
;           Since Hub hardware has no knowledge of this event, firmware must
;           initiate the resume (K) signaling upstream and downstream to all
;           enabled (unsuspended) ports.
;           A call to SEND_GLOBAL_RESUME is also made in this case.
;------------------------------------------------------------------------
THS_SUSPEND_WAIT:
    MOV     A, [HUB_SUSPEND_COUNT]      ; get hub suspend counter
    CMP     A, 0                        ; if 0, atcivity detected, get out
    JZ      TASK_HUB_SUSPEND_END
    CMP     A, HUB_SUSPEND_LIMIT        ; if = LIMIT, go on
    JNZ     THS_SUSPEND_WAIT            ; else wait for limit or activity

;Suspend conditions----------------------
THS_SUSPEND_CASE_1:                     ; pending resume
    MOV     A, [STATE_RESUME_PENDING]   ; get pending resume state
    CMP     A, RESUME_START             ; if START or OVERRIDE
    JNC     THS_RESUME_ENABLED_PORTS    ; don't suspend

THS_SUSPEND_CASE_2:                     ; test C_PORT_SUSPEND bits
    MOV     A, 0                        ; A = port number
    MOV     X, HUB_wPortChange0_LSB     ; point X to HUB_wPortChange0_LSB

THS_S2_LOOP:
    PUSH    A                           ; save port
    MOV     A, [X+0]                    ; get HUB_wPortChangeX_LSB
    AND     A, MASK_PORT_SUSPEND        ; mask C_PORT_SUSPEND bit
    JNZ     THS_SUSPEND_SKIP            ; if not 0, bit set
    MOV     A, X                        ; X += 4 (next port change)
    ADD     A, 4
    MOV     X, A
    POP     A                           ; restore port
    INC     A                           ; bump port count
    CMP     A, MAX_PORTS                ; last port?, fall thru if done, none set
    JC      THS_S2_LOOP                 ; no, test next port

THS_SUSPEND_CASE_3:                     ; if remote wakeup, test C_PORT_CONNECTION bits
    test_flag   FLAG_REMOTE_WAKEUP      ; test if DEVICE_REMOTE_WAKEUP was enabled
    JNZ      THS_SUSPEND                ; if not, suspend!!!

    MOV     A, 0                        ; A = port number
    MOV     X, HUB_wPortChange0_LSB     ; point X to HUB_wPortChange0_LSB

THS_S3_LOOP:
    PUSH    A                           ; save port
    MOV     A, [X+0]                    ; get HUB_wPortChangeX_LSB
    AND     A, MASK_PORT_CONNECTION     ; mask C_PORT_CONNECTION bit
    JNZ     THS_SUSPEND_SKIP            ; if not 0, bit set
    MOV     A, X                        ; X += 4 (next port change)
    ADD     A, 4
    MOV     X, A
    POP     A                           ; restore port
    INC     A                           ; bump port count
     CMP     A, MAX_PORTS                ; last port?, fall thru if done, none set
    JC      THS_S3_LOOP                 ; not, test next port
    JMP     THS_SUSPEND

THS_SUSPEND_SKIP:                       ; C_PORT_SUSPEND or C_PORT_CONNECTION bit set
    POP     A                           ; restore stack
    JMP     THS_RESUME_ENABLED_PORTS    ; do not suspend

THS_SUSPEND:
;(SIMULATE_SUSPEND)**************************
;    read    REG_PORT_ENABLE                     ; get enable register
;    MOV     [BITMAP_ENABLED_UNSUSPENDED], A     ; save
;    read    REG_PORT_SUSPEND                    ; read suspend register
;    PUSH    A                                   ; save suspend bits
;    CPL                                         ; and complement
;    AND     A, [BITMAP_ENABLED_UNSUSPENDED]     ; mask all enabled unsuspended ports
;    MOV     [BITMAP_ENABLED_UNSUSPENDED], A     ; save
;    POP     A                                   ; restore suspend bits
;    OR      A, [BITMAP_ENABLED_UNSUSPENDED]     ; suspend all enabled unsuspended ports
;    write_A REG_PORT_SUSPEND                    ; write to suspend register
;    MOV     A, [BITMAP_ENABLED_UNSUSPENDED]     ; get enabled unsuspended bits
;    CPL                                         ; complement
;    MOV     [BITMAP_ENABLED_UNSUSPENDED], A     ; save back as a mask after resume
;********************************************

    clr_flag    FLAG_HUB_SUSPEND
    MOV     A, [STATE_HUB]              ; get hub state
    OR      A, MASK_HUB_SUSPEND         ; set suspend mask bit
    MOV     [STATE_HUB], A              ; save
    feed_dog                            ; late night snack

    hid_suspend

    disable_1ms_timer

    suspend                             ; !!!!!!!! sleep !!!!!!!!
;(SIMULATE_SUSPEND)********************
;    CALL    SIMULATE_SUSPEND           ; (see SIMULATE.ASM)
;**************************************

    enable_1ms_timer
    feed_dog                            ; morning snack
    MOV     A, [STATE_HUB]              ; get STATE_HUB
    AND     A, ~MASK_HUB_SUSPEND        ; clear suspend mask bit
    MOV     [STATE_HUB], A              ; save


    hid_resume
    ;clear debounce counter for overcurrent.  After a resume, we want to start
    ;debouncing from scratch.

    MOV     A,0
    MOV     [HUB_OVERCURRENT_COUNT],A   ; clear debounce counter

; Resume conditions --------------------
THS_RESUME_CASE_3:
    test_flag   FLAG_PORT_CHANGE            ; connect/disconnect event
    JZ      THS_CONNECT_INITIATED
THS_RESUME_CASE_2:

    test_state  STATE_RESUME_PENDING, RESUME_START ; resume (K) downstream
    JZ      THS_PORT_INITIATED                     ; suspended

    test_hid_resume

; Test for Resume (K) on US (root) port, or resume K enabled DS port
; HW reflects (K) to all enabled DS ports. Terminate by writing to control reg's
THS_RESUME_CASE_1:


    MOV     A, 0                        ; set port control to not forcing
    write_A REG_PORT_CTRL_LOW
    write_A REG_PORT_CTRL_UP
    JMP     TASK_HUB_SUSPEND_END        ; done

; Test for Resume (K) on DS suspended port (device remote wakeup)
; .....reflect (K) US and all DS enabled (unsuspended) ports (>1ms)
THS_PORT_INITIATED:
    MOV     A, RESUME_COMPLETE          ; set resume state to complete
    MOV     [STATE_RESUME_PENDING], A
    MOV     A, [SHADOW_PORT_RESUME]     ; get resume bitmap

;(SIMULATE_SUSPEND)***************
;    AND     A, [BITMAP_ENABLED_UNSUSPENDED]; mask off enabled unsuspended ports
;**************************************

    CMP     A, 0                        ; if bitmap = 0
    JZ      THS_ENABLED_UNSUSPENDED     ; an enabled port initiated resume
    CALL    PORT_FROM_BITMAP            ; else a suspended port initiated resume, get port resuming
    SWAP    A, X                        ; A = new bitmap, X = port
    MOV     [SHADOW_PORT_RESUME], A     ; save bitmap
    SWAP    A, X                        ; A = suspended port initiating resume
    JMP     THS_RESUME_SUSPENDED_PORT
THS_ENABLED_UNSUSPENDED:                ; an enabled port initiated resume
    MOV     [SHADOW_PORT_RESUME], A
    JMP     THS_RESUME_ENABLED_PORTS

; If Hub is DEVICE_REMOTE_WAKEUP, test for connect/disconnect on DS port
; Firmware must control resume (K) signaling US and all enabled (unsuspended)
; DS ports for > 1ms

THS_CONNECT_INITIATED:
;    clr_port0_bit MASK_BIT_4
	CALL    TASK_PORT_CHANGE            ; handle connect/disconnect event
    test_flag   FLAG_REMOTE_WAKEUP      ; test if DEVICE_REMOTE_WAKEUP was enabled
  	jnz      THS_SUSPEND                ; if not, suspend!!!
THS_RESUME_ENABLED_PORTS:
    MOV     A, FFh                      ; set invalid suspend port
THS_RESUME_SUSPENDED_PORT:
    CALL    SEND_GLOBAL_RESUME          ; send resume signaling
;    clr_port0_bit MASK_BIT_4

TASK_HUB_SUSPEND_END:
    clr_flag FLAG_HUB_SUSPEND

    RET